/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */
package lidaws3d.modules;

import edu.memphis.ccrg.lida.environment.EnvironmentImpl;
import edu.memphis.ccrg.lida.framework.tasks.FrameworkTaskImpl;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;
import ws3dproxy.CommandExecException;
import ws3dproxy.CommandUtility;
import ws3dproxy.WS3DProxy;
import ws3dproxy.model.Creature;
import ws3dproxy.model.Leaflet;
import ws3dproxy.model.Thing;
import ws3dproxy.model.World;
import ws3dproxy.util.Constants;

/**
 *
 * @author jucah
 */
public class Environment extends EnvironmentImpl {

    private static final Logger logger = Logger.getLogger(Environment.class.getCanonicalName());
    private static final int DEFAULT_TICKS_PER_RUN = 100;
    private int ticksPerRun;
    private WS3DProxy proxy;
    private Creature creature;
    private World world;
    private Thing food;
    private Thing jewel;
    private List<Thing> thingsAhead;
    private Thing leafletJewel;
    private Map<Long, Integer> completeLeaflets;
    private boolean firstUpdate;
    private boolean done;
    private double score;
    private String currentAction;

    public Environment() {
        this.ticksPerRun = DEFAULT_TICKS_PER_RUN;
        this.score = 0.0;
        this.creature = null;
        this.proxy = new WS3DProxy();
        this.world = null;
        this.creature = null;
        this.food = null;
        this.jewel = null;
        this.thingsAhead = new ArrayList<>();
        this.leafletJewel = null;
        this.firstUpdate = true;
        this.done = false;
        this.completeLeaflets = new HashMap<>();
        this.currentAction = "none";
    }

    @Override
    public void init() {
        super.init();
        ticksPerRun = (Integer) getParam("environment.ticksPerRun", DEFAULT_TICKS_PER_RUN);
        taskSpawner.addTask(new BackgroundTask(ticksPerRun));
        try {
            creature = proxy.createCreature(100, 100, 0);
            world = proxy.getWorld();
            creature.start();
            logger.log(Level.INFO, "{0} started", creature.getName());
            if (World.getNumberOfCreatures() == 1) {
                World.grow(1);
            }
            Thread.sleep(3000);
            creature.updateState();
            logger.log(Level.INFO, "[init] fuel: {0}; leaflets: {1}; thingsInVision: {2}", new Object[]{creature.getFuel(), creature.getLeaflets().size(), creature.getThingsInVision().size()});
        } catch (CommandExecException | InterruptedException ex) {
            logger.log(Level.SEVERE, "Error initializing environment.", ex);
        }
    }

    private class BackgroundTask extends FrameworkTaskImpl {

        public BackgroundTask(int ticksPerRun) {
            super(ticksPerRun);
        }

        @Override
        protected void runThisFrameworkTask() {
            updateState();
            performAction(currentAction);
        }
    }

    @Override
    public void resetState() {
        currentAction = "none";
    }

    @Override
    public Object getState(Map<String, ?> params) {
        logger.log(Level.INFO, "[getState] Requested.");
        //updateState();
        Object requestedObject = null;
        String mode = (String) params.get("mode");
        switch (mode) {
            case "done":
                requestedObject = done;
                break;
            case "energyLevel":
                requestedObject = creature.getFuel();
                break;
            case "food":
                requestedObject = food;
                break;
            case "jewel":
                requestedObject = jewel;
                break;
            case "thingsAhead":
                requestedObject = thingsAhead;
                break;
            case "leafletJewel":
                requestedObject = leafletJewel;
                break;
            case "completeLeaflets":
                requestedObject = completeLeaflets;
                break;
            default:
                logger.log(Level.INFO, "Unknown mode requested: {0}.", mode);
                break;
        }
        logger.log(Level.INFO, "[getState] Mode requested: {0}; object returned: {1}.", new Object[]{mode, requestedObject});
        return requestedObject;
    }

    public void updateState() {
        creature.updateState();
        logger.log(Level.INFO, "[updateState] fuel: {0}; leaflets: {1}; thingsInVision: {2}",
                new Object[]{creature.getFuel(), creature.getLeaflets().size(), creature.getThingsInVision().size()});
        food = null;
        jewel = null;
        leafletJewel = null;
        thingsAhead.clear();
        for (Thing thing : creature.getThingsInVision()) {
            // Constants.OFFSET is the distance from creature path that a thing can be getted
            if (thing.getDistanceToCreature() <= Constants.OFFSET) {
                thingsAhead.add(thing);
            } else {
                if (thing.category == Constants.categoryJEWEL) {
                    if (leafletJewel == null && creature.hasInLeaflet(thing)) {
                        leafletJewel = thing;
                    } else if (jewel != null) {
                        jewel = thing;
                    }
                } else if (food == null
                        && (thing.category == Constants.categoryFOOD
                        || thing.category == Constants.categoryPFOOD
                        || thing.category == Constants.categoryNPFOOD)) {
                    food = thing;
                }
            }
        }
        completeLeaflets.clear();
        for (Leaflet l : creature.getLeaflets()) {
            if (l.isCompleteLeaflet()) {
                completeLeaflets.put(l.getID(), l.getPayment());
            }
        }

        if (!firstUpdate && creature.getLeaflets().isEmpty()) {
            done = true;
        }
        firstUpdate = false;
        logger.log(Level.INFO, "[updateState] fuel: {0}; hasLeaflet: {1}; leaflets: {2}; thingsInVision: {3}",
                new Object[]{creature.getFuel(), creature.getIfHasLeaflet(), creature.getLeaflets().size(), creature.getThingsInVision().size()});
        logger.log(Level.INFO, "[updateState] done: {0}; leafletJewel: {1}; food: {2}; thingsAhead: {3} completeLeaflets: {4}",
                new Object[]{done, leafletJewel, food, thingsAhead.size(), completeLeaflets});
    }

    @Override
    public void processAction(Object action) {
        String actionName = (String) action;
        logger.log(Level.INFO, "[processAction] Processing action: {0}.", actionName);
        if (!done) {
            currentAction = actionName.substring(actionName.indexOf(".") + 1);
            //performAction(currentAction);
        }
    }

    private void performAction(String currentAction) {
        logger.log(Level.INFO, "[performAction] Processing action: {0}.", currentAction);
        try {
            switch (currentAction) {
                case "none":
                // rotate
                case "rotate":
                    CommandUtility.sendSetTurn(creature.getIndexID(), 0.0, 0.0, 2.0);
                    logger.log(Level.INFO, "[performAction] Rotating.");
                    break;
                case "goto":        // go to the closest thing (food if low energy) leafletJewel if (enough energy), if any
                    if (creature.getFuel() > 400.0 && leafletJewel != null && thingsAhead.isEmpty()) {
                        CommandUtility.sendGoTo(creature.getIndexID(), 1.0, 1.0, leafletJewel.getX1(), leafletJewel.getY1());
                        logger.log(Level.INFO, "[performAction] Going to {0} at ({1};{2}).", new Object[]{leafletJewel.getName(), leafletJewel.getX1(), leafletJewel.getY1()});
                    } else if (food != null) {
                        CommandUtility.sendGoTo(creature.getIndexID(), 1.0, 1.0, food.getX1(), food.getY1());
                        logger.log(Level.INFO, "[performAction] Going to {0} at ({1};{2}).", new Object[]{food.getName(), food.getX1(), food.getY1()});
                    }
                    break;
                case "get":         //get (sack or eat) all things close to creature
                    CommandUtility.sendSetTurn(creature.getIndexID(), 0.0, 0.0, 0.0);
                    if (thingsAhead != null) {
                        for (Thing ta : thingsAhead) {
                            Object[] logParams = new Object[]{ta.getName(), ta.getX1(), ta.getY1()};
                            boolean isFood = ta.category == Constants.categoryFOOD || ta.category == Constants.categoryNPFOOD || ta.category == Constants.categoryPFOOD;
                            if (ta.category == Constants.categoryJEWEL) {
                                creature.putInSack(ta.getName());
                                logger.log(Level.INFO, "[performAction] Getting {0} at ({1};{2}).", logParams);
                            } else if (isFood) {
                                creature.eatIt(ta.getName());
                                logger.log(Level.INFO, "[performAction] Eating {0} at ({1};{2}).", logParams);
                            }
                        }
                    }
                    this.resetState();
                    break;
                case "deliver":     // deliver all complete leaflets
                    for (Iterator<Long> iter = completeLeaflets.keySet().iterator(); iter.hasNext();) {
                        Long leafletID = iter.next();
                        int newScore = completeLeaflets.get(leafletID);
                        creature.deliverLeaflet(String.valueOf(leafletID));
                        logger.log(Level.INFO, "[performAction] Delivering leaflet {0} with score: {1}.", new Object[]{String.valueOf(leafletID), newScore});
                        score = score + newScore;
                        completeLeaflets.remove(leafletID);
                    }
                    break;
                case "stop":        // stop the creature and log results
                    creature.stop();
                    logger.log(Level.INFO, "{0} stopped.\n", new Object[]{creature.getName(), score});
                    logger.log(Level.INFO, "Done! {0} has completed all his leaflets obtaining a score of {1} points.\n", new Object[]{creature.getName(), score});
                    break;
                default:
                    logger.log(Level.INFO, "[performAction] Unknown action requested: {0}.", currentAction);
                    break;
            }
        } catch (CommandExecException ex) {
            logger.log(Level.SEVERE, "[performAction] Error processing action.", ex);
        }
    }
}
